from .curve_loop import CurveLoop


class Surface:
    """
    Generates a Surface from a CurveLoop.

    Parameters
    ----------
    curve_loop : Object
        CurveLoop object that contains all the Line objects for the loop construction.

    Notes
    -----
    With the built-in kernel, the first line loop should be composed of either three or
    four elementary lines.

    With the built-in kernel, the optional In Sphere argument forces the surface to be a
    spherical patch (the extra parameter gives the identification number of the center
    of the sphere).
    """

    dim = 2

    def __init__(self, env, curve_loop):
        assert isinstance(curve_loop, CurveLoop)
        self.curve_loop = curve_loop
        self.num_edges = len(curve_loop)
        self._ID = env.addSurfaceFilling([self.curve_loop._ID])

    def __repr__(self):
        return f"<pygmsh Surface object, ID {self._ID}>"



from math import pi

import gmsh


class Cone:
    """
    Creates a cone.

    center : array-like[3]
        The 3 coordinates of the center of the first circular face.
    axis : array-like[3]
        The 3 components of the vector defining its axis.
    radius0 : float
        Radius of the first circle.
    radius1 : float
        Radius of the second circle.
    angle : float
        Angular opening of the the Cone.
    """

    dim = 3

    def __init__(self, center, axis, radius0, radius1, angle=2 * pi):
        assert len(center) == 3
        assert len(axis) == 3

        self.center = center
        self.axis = axis
        self.radius0 = radius0
        self.radius1 = radius1

        self._ID = gmsh.model.occ.addCone(*center, *axis, radius0, radius1, angle=angle)
        self.dim_tags = [(3, self._ID)]

    def __repr__(self):
        return f"<pygmsh Cone object, ID {self._ID}>"



import gmsh


class Disk:
    """
    Creates a disk.

    Parameters
    ----------
    x0 : array-like[3]
        The 3 coordinates of the center of the disk face.
    radius0 : float
        Radius value of the disk.
    radius1 : float
        Radius along Y, leading to an ellipse.
    """

    dim = 2

    def __init__(self, x0, radius0, radius1=None):
        if len(x0) == 2:
            x0 = [x0[0], x0[1], 0.0]
        assert len(x0) == 3

        if radius1 is None:
            radius1 = radius0
        assert radius0 >= radius1

        self.x0 = x0
        self.radius0 = radius0
        self.radius1 = radius1

        self._ID = gmsh.model.occ.addDisk(*x0, radius0, radius1)
        self.dim_tag = (self.dim, self._ID)
        self.dim_tags = [self.dim_tag]

    def __repr__(self):
        return f"<pygmsh Disk object, ID {self._ID}>"



import gmsh
import meshio
import numpy


def rotation_matrix(u, theta):
    """Return matrix that implements the rotation around the vector :math:`u`
    by the angle :math:`\\theta`, cf.
    https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle.

    :param u: rotation vector
    :param theta: rotation angle
    """
    assert numpy.isclose(numpy.inner(u, u), 1.0), "the rotation axis must be unitary"

    # Cross-product matrix.
    cpm = numpy.array([[0.0, -u[2], u[1]], [u[2], 0.0, -u[0]], [-u[1], u[0], 0.0]])
    c = numpy.cos(theta)
    s = numpy.sin(theta)
    R = numpy.eye(3) * c + s * cpm + (1.0 - c) * numpy.outer(u, u)
    return R


def orient_lines(lines):
    """Given a sequence of unordered and unoriented lines defining a closed polygon,
    returns a reordered list of reoriented lines of that polygon.

    :param lines: a sequence of lines defining a closed polygon
    """
    # Categorise graph edges by their vertex pair ids
    point_pair_ids = numpy.array(
        [[line.points[0]._ID, line.points[1]._ID] for line in lines]
    )

    # Indices of reordering
    order = numpy.arange(len(point_pair_ids), dtype=int)
    # Compute orientations where oriented[j] == False requires edge j to be reversed
    oriented = numpy.array([True] * len(point_pair_ids), dtype=numpy.bool)

    for j in range(1, len(point_pair_ids)):
        out = point_pair_ids[j - 1, 1]  # edge out from vertex
        inn = point_pair_ids[j:, 0]  # candidates for edge into vertices
        wh = numpy.where(inn == out)[0] + j
        if len(wh) == 0:
            # look for candidates in those which are not correctly oriented
            inn = point_pair_ids[j:, 1]
            wh = numpy.where(inn == out)[0] + j
            # reorient remaining edges
            point_pair_ids[j:] = numpy.flip(point_pair_ids[j:], axis=1)
            oriented[j:] ^= True

        # reorder
        point_pair_ids[[j, wh[0]]] = point_pair_ids[[wh[0], j]]
        order[[j, wh[0]]] = order[[wh[0], j]]

    # Reconstruct an ordered and oriented line loop
    lines = [lines[o] for o in order]
    lines = [lines[j] if oriented[j] else -lines[j] for j in range(len(oriented))]

    return lines


def extract_to_meshio():
    # extract point coords
    idx, points, _ = gmsh.model.mesh.getNodes()
    points = points.reshape(-1, 3)
    idx -= 1
    srt = numpy.argsort(idx)
    assert numpy.all(idx[srt] == numpy.arange(len(idx)))
    points = points[srt]

    # extract cells
    elem_types, elem_tags, node_tags = gmsh.model.mesh.getElements()
    cells = []
    for elem_type, node_tags in zip(elem_types, node_tags):
        # `elementName', `dim', `order', `numNodes', `localNodeCoord',
        # `numPrimaryNodes'
        num_nodes_per_cell = gmsh.model.mesh.getElementProperties(elem_type)[3]
        meshio.gmsh.gmsh_to_meshio_type
        cells.append(
            meshio.CellBlock(
                meshio.gmsh.gmsh_to_meshio_type[elem_type],
                node_tags.reshape(-1, num_nodes_per_cell) - 1,
            )
        )

    cell_sets = {}
    for dim, tag in gmsh.model.getPhysicalGroups():
        name = gmsh.model.getPhysicalName(dim, tag)
        cell_sets[name] = [[] for _ in range(len(cells))]
        for e in gmsh.model.getEntitiesForPhysicalGroup(dim, tag):
            # TODO node_tags?
            # elem_types, elem_tags, node_tags
            elem_types, elem_tags, _ = gmsh.model.mesh.getElements(dim, e)
            assert len(elem_types) == len(elem_tags)
            assert len(elem_types) == 1
            elem_type = elem_types[0]
            elem_tags = elem_tags[0]

            meshio_cell_type = meshio.gmsh.gmsh_to_meshio_type[elem_type]
            # make sure that the cell type appears only once in the cell list
            # -- for now
            idx = []
            for k, cell_block in enumerate(cells):
                if cell_block.type == meshio_cell_type:
                    idx.append(k)
            assert len(idx) == 1
            idx = idx[0]
            cell_sets[name][idx].append(elem_tags - 1)

        cell_sets[name] = [
            (None if len(idcs) == 0 else numpy.concatenate(idcs))
            for idcs in cell_sets[name]
        ]

    # make meshio mesh
    return meshio.Mesh(points, cells, cell_sets=cell_sets)



from math import pi

import gmsh


class Cylinder:
    """
    Creates a cylinder.

    Parameters
    ----------
    x0 : array-like[3]
        The 3 coordinates of the center of the first circular face.
    axis : array-like[3]
        The 3 components of the vector defining its axis.
    radius : float
        Radius value of the cylinder.
    angle : float
        Angular opening of the cylinder.
    """

    def __init__(self, x0, axis, radius, angle=2 * pi):
        assert len(x0) == 3
        assert len(axis) == 3

        self.x0 = x0
        self.axis = axis
        self.radius = radius
        self.angle = angle

        self._ID = gmsh.model.occ.addCylinder(*x0, *axis, radius, angle=angle)
        self.dim_tag = (3, self._ID)
        self.dim_tags = [self.dim_tag]

    def __repr__(self):
        return f"<pygmsh Cylinder object, ID {self._ID}>"



import gmsh


class Rectangle:
    """
    Creates a rectangle.

    x0 : array-like[3]
         The 3 first expressions define the lower-left corner.
    a : float
        Rectangle width.
    b : float
        Rectangle height.
    corner_radius : float
        Defines a radius to round the rectangle corners.
    """

    dim = 2

    def __init__(self, x0, a, b, corner_radius=None):
        assert len(x0) == 3

        self.x0 = x0
        self.a = a
        self.b = b
        self.corner_radius = corner_radius

        if corner_radius is None:
            corner_radius = 0.0

        self._ID = gmsh.model.occ.addRectangle(*x0, a, b, roundedRadius=corner_radius)
        self.dim_tag = (self.dim, self._ID)
        self.dim_tags = [self.dim_tag]

    def __repr__(self):
        return f"<pygmsh Rectangle object, ID {self._ID}>"



from .line_base import LineBase
from .point import Point


class Spline(LineBase):
    """
    With the built-in geometry kernel this constructs a Catmull-Rom spline.

    Parameters
    ----------
    points : list
        List containing Point objects
    """

    def __init__(self, env, points):
        for c in points:
            assert isinstance(c, Point)
        assert len(points) > 1

        id0 = env.addSpline([c._ID for c in points])
        super().__init__(id0, points)



from . import geo, occ
from .__about__ import __gmsh_version__, __version__
from ._optimize import optimize
from .helpers import orient_lines, rotation_matrix

__all__ = [
    "geo",
    "occ",
    "rotation_matrix",
    "orient_lines",
    "optimize",
    "__version__",
    "__gmsh_version__",
]



import gmsh

from ..helpers import extract_to_meshio
from .bspline import BSpline
from .circle_arc import CircleArc
from .curve_loop import CurveLoop
from .dummy import Dummy
from .ellipse_arc import EllipseArc
from .line import Line
from .plane_surface import PlaneSurface
from .point import Point
from .polygon import Polygon
from .spline import Spline
from .surface import Surface
from .surface_loop import SurfaceLoop
from .volume import Volume


class CommonGeometry:
    """Geometry base class containing all methods that can be shared between built-in
    and occ.
    """

    def __init__(self, env):
        self.env = env
        self._BOOLEAN_ID = 0
        self._ARRAY_ID = 0
        self._FIELD_ID = 0
        self._COMPOUND_ENTITIES = []
        self._RECOMBINE_ENTITIES = []
        self._EMBED_QUEUE = []
        self._TRANSFINITE_CURVE_QUEUE = []
        self._TRANSFINITE_SURFACE_QUEUE = []
        self._TRANSFINITE_VOLUME_QUEUE = []
        self._AFTER_SYNC_QUEUE = []
        self._SIZE_QUEUE = []
        self._PHYSICAL_QUEUE = []
        self._OUTWARD_NORMALS = []

    def __enter__(self):
        gmsh.initialize()
        gmsh.model.add("pygmsh model")
        return self

    def __exit__(self, *a):
        try:
            # Gmsh >= 4.7.0
            # https://gitlab.onelab.info/gmsh/gmsh/-/issues/1036
            gmsh.model.mesh.removeSizeCallback()
        except AttributeError:
            pass
        gmsh.finalize()

    def synchronize(self):
        self.env.synchronize()

    def __repr__(self):
        return "<pygmsh Geometry object>"

    def add_bspline(self, *args, **kwargs):
        return BSpline(self.env, *args, **kwargs)

    def add_circle_arc(self, *args, **kwargs):
        return CircleArc(self.env, *args, **kwargs)

    def add_ellipse_arc(self, *args, **kwargs):
        return EllipseArc(self.env, *args, **kwargs)

    def add_line(self, *args, **kwargs):
        return Line(self.env, *args, **kwargs)

    def add_curve_loop(self, *args, **kwargs):
        return CurveLoop(self.env, *args, **kwargs)

    def add_plane_surface(self, *args, **kwargs):
        return PlaneSurface(self.env, *args, **kwargs)

    def add_point(self, *args, **kwargs):
        return Point(self.env, *args, **kwargs)

    def add_spline(self, *args, **kwargs):
        return Spline(self.env, *args, **kwargs)

    def add_surface(self, *args, **kwargs):
        return Surface(self.env, *args, **kwargs)

    def add_surface_loop(self, *args, **kwargs):
        return SurfaceLoop(self.env, *args, **kwargs)

    def add_volume(self, *args, **kwargs):
        return Volume(self.env, *args, **kwargs)

    def add_polygon(self, *args, **kwargs):
        return Polygon(self, *args, **kwargs)

    def add_physical(self, entities, label=None):
        if not isinstance(entities, list):
            entities = [entities]

        # make sure the dimensionality is the same for all entities
        dim = entities[0].dim
        for e in entities:
            assert e.dim == dim

        if label is not None:
            if not isinstance(label, str):
                raise ValueError(f"Physical label must be string, not {type(label)}.")

        self._PHYSICAL_QUEUE.append((entities, label))

    def set_transfinite_curve(self, curve, num_nodes, mesh_type, coeff):
        assert mesh_type in ["Progression", "Bulk"]
        self._TRANSFINITE_CURVE_QUEUE.append((curve._ID, num_nodes, mesh_type, coeff))

    def set_transfinite_surface(self, surface, arrangement, corner_tags):
        self._TRANSFINITE_SURFACE_QUEUE.append((surface._ID, arrangement, corner_tags))

    def set_transfinite_volume(self, volume, arrangement, corner_tags):
        self._TRANSFINITE_VOLUME_QUEUE.append((volume._ID, corner_tags))

    def set_recombined_surfaces(self, surfaces):
        for i, surface in enumerate(surfaces):
            assert isinstance(
                surface, (PlaneSurface, Surface)
            ), f"item {i} is not a surface"
        self._RECOMBINE_ENTITIES += [s.dim_tags[0] for s in surfaces]

    def extrude(
        self,
        input_entity,
        translation_axis,
        num_layers=None,
        heights=None,
        recombine=False,
    ):
        """Extrusion of any entity along a given translation_axis."""
        if isinstance(num_layers, int):
            num_layers = [num_layers]
        if num_layers is None:
            num_layers = []
            heights = []
        else:
            if heights is None:
                heights = []
            else:
                assert len(num_layers) == len(heights)

        assert len(translation_axis) == 3

        ie_list = input_entity if isinstance(input_entity, list) else [input_entity]

        out_dim_tags = self.env.extrude(
            [e.dim_tag for e in ie_list],
            *translation_axis,
            numElements=num_layers,
            heights=heights,
            recombine=recombine,
        )
        top = Dummy(*out_dim_tags[0])
        extruded = Dummy(*out_dim_tags[1])
        lateral = [Dummy(*e) for e in out_dim_tags[2:]]
        return top, extruded, lateral

    def _revolve(
        self,
        input_entity,
        rotation_axis,
        point_on_axis,
        angle,
        num_layers=None,
        heights=None,
        recombine=False,
    ):
        """Rotation of any entity around a given rotation_axis, about a given angle."""
        if isinstance(num_layers, int):
            num_layers = [num_layers]
        if num_layers is None:
            num_layers = []
            heights = []
        else:
            if heights is None:
                heights = []
            else:
                assert len(num_layers) == len(heights)

        assert len(point_on_axis) == 3
        assert len(rotation_axis) == 3
        out_dim_tags = self.env.revolve(
            input_entity.dim_tags,
            *point_on_axis,
            *rotation_axis,
            angle,
            numElements=num_layers,
            heights=heights,
            recombine=recombine,
        )

        top = Dummy(*out_dim_tags[0])
        extruded = Dummy(*out_dim_tags[1])
        lateral = [Dummy(*e) for e in out_dim_tags[2:]]
        return top, extruded, lateral

    def translate(self, obj, vector):
        """Translates input_entity itself by vector.

        Changes the input object.
        """
        self.env.translate(obj.dim_tags, *vector)

    def rotate(self, obj, point, angle, axis):
        """Rotate input_entity around a given point with a give angle.
           Rotation axis has to be specified.

        Changes the input object.
        """
        self.env.rotate(obj.dim_tags, *point, *axis, angle)

    def copy(self, obj):
        dim_tag = self.env.copy(obj.dim_tags)
        assert len(dim_tag) == 1
        return Dummy(*dim_tag[0])

    def symmetrize(self, obj, coefficients):
        """Transforms all elementary entities symmetrically to a plane. The vector
        should contain four expressions giving the coefficients of the plane's equation.
        """
        self.env.symmetrize(obj.dim_tags, *coefficients)

    def dilate(self, obj, x0, abc):
        self.env.dilate(obj.dim_tags, *x0, *abc)

    def mirror(self, obj, abcd):
        self.env.mirror(obj.dim_tags, *abcd)

    def remove(self, obj, recursive=False):
        self.env.remove(obj.dim_tags, recursive=recursive)

    def in_surface(self, input_entity, surface):
        """Embed the point(s) or curve(s) in the given surface. The surface mesh will
        conform to the mesh of the point(s) or curves(s).
        """
        self._EMBED_QUEUE.append((input_entity, surface))

    def in_volume(self, input_entity, volume):
        """Embed the point(s)/curve(s)/surface(s) in the given volume. The volume mesh
        will conform to the mesh of the input entities.
        """
        self._EMBED_QUEUE.append((input_entity, volume))

    def set_mesh_size_callback(self, fun, ignore_other_mesh_sizes=True):
        gmsh.model.mesh.setSizeCallback(fun)
        #
        # If a mesh size is set from a function, ignore the mesh sizes from the
        # entities.
        #
        # From <http://gmsh.info/doc/texinfo/gmsh.html#t10>:
        # ```
        # To determine the size of mesh elements, Gmsh locally computes the minimum of
        #
        # 1) the size of the model bounding box;
        # 2) if `Mesh.CharacteristicLengthFromPoints' is set, the mesh size specified at
        #    geometrical points;
        # 3) if `Mesh.CharacteristicLengthFromCurvature' is set, the mesh size based on
        #    the curvature and `Mesh.MinimumElementsPerTwoPi';
        # 4) the background mesh field;
        # 5) any per-entity mesh size constraint.
        #
        # This value is then constrained in the interval
        # [`Mesh.CharacteristicLengthMin', `Mesh.CharacteristicLengthMax'] and
        # multiplied by `Mesh.CharacteristicLengthFactor'.  In addition, boundary mesh
        # sizes (on curves or surfaces) are interpolated inside the enclosed entity
        # (surface or volume, respectively) if the option
        # `Mesh.CharacteristicLengthExtendFromBoundary' is set (which is the case by
        # default).
        # ```
        if ignore_other_mesh_sizes:
            gmsh.option.setNumber("Mesh.CharacteristicLengthExtendFromBoundary", 0)
            gmsh.option.setNumber("Mesh.CharacteristicLengthFromPoints", 0)
            gmsh.option.setNumber("Mesh.CharacteristicLengthFromCurvature", 0)

    def generate_mesh(  # noqa: C901
        self,
        dim=3,
        order=None,
        # http://gmsh.info/doc/texinfo/gmsh.html#index-Mesh_002eAlgorithm
        algorithm=None,
        verbose=False,
    ):
        """Return a meshio.Mesh, storing the mesh points, cells, and data, generated by
        Gmsh from the `self`.
        """
        self.synchronize()

        for item in self._AFTER_SYNC_QUEUE:
            item.exec()

        for item, host in self._EMBED_QUEUE:
            gmsh.model.mesh.embed(item.dim, [item._ID], host.dim, host._ID)

        # set compound entities after sync
        for c in self._COMPOUND_ENTITIES:
            gmsh.model.mesh.setCompound(*c)

        for s in self._RECOMBINE_ENTITIES:
            gmsh.model.mesh.setRecombine(*s)

        for t in self._TRANSFINITE_CURVE_QUEUE:
            gmsh.model.mesh.setTransfiniteCurve(*t)

        for t in self._TRANSFINITE_SURFACE_QUEUE:
            gmsh.model.mesh.setTransfiniteSurface(*t)

        for e in self._TRANSFINITE_VOLUME_QUEUE:
            gmsh.model.mesh.setTransfiniteVolume(*e)

        for item, size in self._SIZE_QUEUE:
            gmsh.model.mesh.setSize(
                gmsh.model.getBoundary(item.dim_tags, False, False, True), size
            )

        for entities, label in self._PHYSICAL_QUEUE:
            d = entities[0].dim
            assert all(e.dim == d for e in entities)
            tag = gmsh.model.addPhysicalGroup(d, [e._ID for e in entities])
            if label is not None:
                gmsh.model.setPhysicalName(d, tag, label)

        for entity in self._OUTWARD_NORMALS:
            gmsh.model.mesh.setOutwardOrientation(entity.id)

        if order is not None:
            gmsh.model.mesh.setOrder(order)

        gmsh.option.setNumber("General.Terminal", 1 if verbose else 0)

        # set algorithm
        # http://gmsh.info/doc/texinfo/gmsh.html#index-Mesh_002eAlgorithm
        if algorithm:
            gmsh.option.setNumber("Mesh.Algorithm", algorithm)

        gmsh.model.mesh.generate(dim)

        return extract_to_meshio()



class Point:
    """
    Creates an elementary point.

    x : array-like[3]
        Give the three X, Y and Z coordinates of the
        point in the three-dimensional Euclidean space.
    mesh_size : float
        The prescribed mesh element size at this point.
    """

    dim = 0

    def __init__(self, env, x, mesh_size=None):
        if len(x) == 2:
            x = [x[0], x[1], 0.0]

        assert len(x) == 3
        self.x = x
        args = list(x)
        if mesh_size is not None:
            args.append(mesh_size)
        self._ID = env.addPoint(*args)
        self.dim_tag = (0, self._ID)
        self.dim_tags = [self.dim_tag]

    def __repr__(self):
        X = ", ".join(str(x) for x in self.x)
        return f"<pygmsh Point object, ID {self._ID}, x = [{X}]>"



from .curve_loop import CurveLoop


class PlaneSurface:
    """
    Creates a plane surface.

    Parameters
    ----------
    curve_loop : Object
        Each unique line in the line loop will be used
        for the surface construction.
    holes : list
        List of line loops that represents polygon holes.

    Notes
    -----
    The first line loop defines the exterior boundary of the surface;
    all other line loops define holes in the surface.

    A line loop defining a hole should not have any lines in
    common with the exterior line loop (in which case it is not
    a hole, and the two surfaces should be defined separately).

    Likewise, a line loop defining a hole should not have any lines
    in common with another line loop defining a hole in the same
    surface (in which case the two line loops should be combined).
    """

    dim = 2

    def __init__(self, env, curve_loop, holes=None):
        assert isinstance(curve_loop, CurveLoop)
        self.curve_loop = curve_loop

        if holes is None:
            holes = []

        # The input holes are either line loops or entities that contain line loops
        # (like polygons).
        self.holes = [h if isinstance(h, CurveLoop) else h.curve_loop for h in holes]

        self.num_edges = len(self.curve_loop) + sum(len(h) for h in self.holes)

        curve_loops = [self.curve_loop] + self.holes
        self._ID = env.addPlaneSurface([ll._ID for ll in curve_loops])
        self.dim_tag = (2, self._ID)
        self.dim_tags = [self.dim_tag]

    def __repr__(self):
        return (
            "<pygmsh PlaneSurface object (OCC), "
            f"ID {self._ID}, curve loop {self.curve_loop._ID}>"
        )



import math

import gmsh
import numpy

from .. import common
from .dummy import Dummy


class Circle:
    def __init__(
        self,
        x0,
        radius,
        R,
        compound,
        num_sections,
        holes,
        curve_loop,
        plane_surface,
        mesh_size=None,
    ):
        self.x0 = x0
        self.radius = radius
        self.mesh_size = mesh_size
        self.R = R
        self.compound = compound
        self.num_sections = num_sections
        self.holes = holes
        self.curve_loop = curve_loop
        self.plane_surface = plane_surface


class Geometry(common.CommonGeometry):
    def __init__(self):
        super().__init__(gmsh.model.geo)

    def revolve(self, *args, **kwargs):
        if len(args) >= 4:
            angle = args[3]
        else:
            assert "angle" in kwargs
            angle = kwargs["angle"]

        assert angle < math.pi
        return super()._revolve(*args, **kwargs)

    def twist(
        self,
        input_entity,
        translation_axis,
        rotation_axis,
        point_on_axis,
        angle,
        num_layers=None,
        heights=None,
        recombine=False,
    ):
        """Twist (translation + rotation) of any entity along a given translation_axis,
        around a given rotation_axis, about a given angle.
        """
        if isinstance(num_layers, int):
            num_layers = [num_layers]
        if num_layers is None:
            num_layers = []
            heights = []
        else:
            if heights is None:
                heights = []
            else:
                assert len(num_layers) == len(heights)

        assert len(point_on_axis) == 3
        assert len(rotation_axis) == 3
        assert len(translation_axis) == 3
        assert angle < math.pi
        out_dim_tags = self.env.twist(
            input_entity.dim_tags,
            *point_on_axis,
            *translation_axis,
            *rotation_axis,
            angle,
            numElements=num_layers,
            heights=heights,
            recombine=recombine,
        )
        top = Dummy(*out_dim_tags[0])
        extruded = Dummy(*out_dim_tags[1])
        lateral = [Dummy(*e) for e in out_dim_tags[2:]]
        return top, extruded, lateral

    def add_circle(
        self,
        x0,
        radius,
        mesh_size=None,
        R=None,
        compound=False,
        num_sections=3,
        holes=None,
        make_surface=True,
    ):
        """Add circle in the :math:`x`-:math:`y`-plane."""
        if holes is None:
            holes = []
        else:
            assert make_surface

        # Define points that make the circle (midpoint and the four cardinal
        # directions).
        X = numpy.zeros((num_sections + 1, len(x0)))
        if num_sections == 4:
            # For accuracy, the points are provided explicitly.
            X[1:, [0, 1]] = numpy.array(
                [[radius, 0.0], [0.0, radius], [-radius, 0.0], [0.0, -radius]]
            )
        else:
            X[1:, [0, 1]] = numpy.array(
                [
                    [
                        radius * numpy.cos(2 * numpy.pi * k / num_sections),
                        radius * numpy.sin(2 * numpy.pi * k / num_sections),
                    ]
                    for k in range(num_sections)
                ]
            )

        if R is not None:
            assert numpy.allclose(
                abs(numpy.linalg.eigvals(R)), numpy.ones(X.shape[1])
            ), "The transformation matrix doesn't preserve circles; at least one eigenvalue lies off the unit circle."
            X = numpy.dot(X, R.T)

        X += x0

        # Add Gmsh Points.
        p = [self.add_point(x, mesh_size=mesh_size) for x in X]

        # Define the circle arcs.
        arcs = [
            self.add_circle_arc(p[k], p[0], p[k + 1]) for k in range(1, len(p) - 1)
        ] + [self.add_circle_arc(p[-1], p[0], p[1])]

        if compound:
            self._COMPOUND_ENTITIES.append((1, [arc._ID for arc in arcs]))

        curve_loop = self.add_curve_loop(arcs)

        if make_surface:
            plane_surface = self.add_plane_surface(curve_loop, holes)
            if compound:
                self._COMPOUND_ENTITIES.append((2, [plane_surface._ID]))
        else:
            plane_surface = None

        return Circle(
            x0,
            radius,
            R,
            compound,
            num_sections,
            holes,
            curve_loop,
            plane_surface,
            mesh_size=mesh_size,
        )

    def add_boundary_layer(self, *args, **kwargs):
        layer = BoundaryLayer(*args, **kwargs)
        self._AFTER_SYNC_QUEUE.append(layer)
        return layer

    def set_background_mesh(self, *args, **kwargs):
        setter = SetBackgroundMesh(*args, **kwargs)
        self._AFTER_SYNC_QUEUE.append(setter)

    def add_rectangle(
        self, xmin, xmax, ymin, ymax, z, mesh_size=None, holes=None, make_surface=True
    ):
        return self.add_polygon(
            [[xmin, ymin, z], [xmax, ymin, z], [xmax, ymax, z], [xmin, ymax, z]],
            mesh_size=mesh_size,
            holes=holes,
            make_surface=make_surface,
        )

    def add_ellipsoid(self, x0, radii, mesh_size=None, with_volume=True, holes=None):
        """Creates an ellipsoid with radii around a given midpoint
        :math:`x_0`.
        """
        if holes is None:
            holes = []

        if holes:
            assert with_volume

        # Add points.
        p = [
            self.add_point(x0, mesh_size=mesh_size),
            self.add_point([x0[0] + radii[0], x0[1], x0[2]], mesh_size=mesh_size),
            self.add_point([x0[0], x0[1] + radii[1], x0[2]], mesh_size=mesh_size),
            self.add_point([x0[0], x0[1], x0[2] + radii[2]], mesh_size=mesh_size),
            self.add_point([x0[0] - radii[0], x0[1], x0[2]], mesh_size=mesh_size),
            self.add_point([x0[0], x0[1] - radii[1], x0[2]], mesh_size=mesh_size),
            self.add_point([x0[0], x0[1], x0[2] - radii[2]], mesh_size=mesh_size),
        ]
        # Add skeleton.
        # Alternative for circles:
        # `self.add_circle_arc(a, b, c)`
        c = [
            self.add_ellipse_arc(p[1], p[0], p[6], p[6]),
            self.add_ellipse_arc(p[6], p[0], p[4], p[4]),
            self.add_ellipse_arc(p[4], p[0], p[3], p[3]),
            self.add_ellipse_arc(p[3], p[0], p[1], p[1]),
            self.add_ellipse_arc(p[1], p[0], p[2], p[2]),
            self.add_ellipse_arc(p[2], p[0], p[4], p[4]),
            self.add_ellipse_arc(p[4], p[0], p[5], p[5]),
            self.add_ellipse_arc(p[5], p[0], p[1], p[1]),
            self.add_ellipse_arc(p[6], p[0], p[2], p[2]),
            self.add_ellipse_arc(p[2], p[0], p[3], p[3]),
            self.add_ellipse_arc(p[3], p[0], p[5], p[5]),
            self.add_ellipse_arc(p[5], p[0], p[6], p[6]),
        ]

        # Add surfaces (1/8th of the ball surface).
        # Make sure the loops are oriented outwards!
        ll = [
            # one half
            self.add_curve_loop([c[4], c[9], c[3]]),
            self.add_curve_loop([c[8], -c[4], c[0]]),
            self.add_curve_loop([-c[9], c[5], c[2]]),
            self.add_curve_loop([-c[5], -c[8], c[1]]),
            # the other half
            self.add_curve_loop([c[7], -c[3], c[10]]),
            self.add_curve_loop([c[11], -c[0], -c[7]]),
            self.add_curve_loop([-c[10], -c[2], c[6]]),
            self.add_curve_loop([-c[1], -c[11], -c[6]]),
        ]

        # Create a surface for each line loop.
        s = [self.add_surface(l) for l in ll]

        # Combine the surfaces to avoid seams
        # <https://gitlab.onelab.info/gmsh/gmsh/issues/507>
        # Cannot enable those yet, <https://gitlab.onelab.info/gmsh/gmsh/-/issues/995>
        self._COMPOUND_ENTITIES.append((2, [surf._ID for surf in s[:4]]))
        self._COMPOUND_ENTITIES.append((2, [surf._ID for surf in s[4:]]))

        # Create the surface loop.
        surface_loop = self.add_surface_loop(s)
        # if holes:
        #     # Create an array of surface loops; the first entry is the outer
        #     # surface loop, the following ones are holes.
        #     surface_loop = self.add_array([surface_loop] + holes)
        # Create volume.
        volume = self.add_volume(surface_loop, holes) if with_volume else None

        class Ellipsoid:
            dim = 3

            def __init__(self, x0, radii, surface_loop, volume, mesh_size=None):
                self.x0 = x0
                self.mesh_size = mesh_size
                self.radii = radii
                self.surface_loop = surface_loop
                self.volume = volume
                return

        return Ellipsoid(x0, radii, surface_loop, volume, mesh_size=mesh_size)

    def add_ball(self, x0, radius, **kwargs):
        return self.add_ellipsoid(x0, [radius, radius, radius], **kwargs)

    def add_box(
        self, x0, x1, y0, y1, z0, z1, mesh_size=None, with_volume=True, holes=None
    ):
        if holes is None:
            holes = []

        if holes:
            assert with_volume

        # Define corner points.
        p = [
            self.add_point([x1, y1, z1], mesh_size=mesh_size),
            self.add_point([x1, y1, z0], mesh_size=mesh_size),
            self.add_point([x1, y0, z1], mesh_size=mesh_size),
            self.add_point([x1, y0, z0], mesh_size=mesh_size),
            self.add_point([x0, y1, z1], mesh_size=mesh_size),
            self.add_point([x0, y1, z0], mesh_size=mesh_size),
            self.add_point([x0, y0, z1], mesh_size=mesh_size),
            self.add_point([x0, y0, z0], mesh_size=mesh_size),
        ]
        # Define edges.
        e = [
            self.add_line(p[0], p[1]),
            self.add_line(p[0], p[2]),
            self.add_line(p[0], p[4]),
            self.add_line(p[1], p[3]),
            self.add_line(p[1], p[5]),
            self.add_line(p[2], p[3]),
            self.add_line(p[2], p[6]),
            self.add_line(p[3], p[7]),
            self.add_line(p[4], p[5]),
            self.add_line(p[4], p[6]),
            self.add_line(p[5], p[7]),
            self.add_line(p[6], p[7]),
        ]

        # Define the six line loops.
        ll = [
            self.add_curve_loop([e[0], e[3], -e[5], -e[1]]),
            self.add_curve_loop([e[0], e[4], -e[8], -e[2]]),
            self.add_curve_loop([e[1], e[6], -e[9], -e[2]]),
            self.add_curve_loop([e[3], e[7], -e[10], -e[4]]),
            self.add_curve_loop([e[5], e[7], -e[11], -e[6]]),
            self.add_curve_loop([e[8], e[10], -e[11], -e[9]]),
        ]

        # Create a surface for each line loop.
        s = [self.add_surface(l) for l in ll]
        # Create the surface loop.
        surface_loop = self.add_surface_loop(s)

        # Create volume
        vol = self.add_volume(surface_loop, holes) if with_volume else None

        class Box:
            def __init__(
                self, x0, x1, y0, y1, z0, z1, surface_loop, volume, mesh_size=None
            ):
                self.x0 = x0
                self.x1 = x1
                self.y0 = y0
                self.y1 = y1
                self.z0 = z0
                self.z1 = z1
                self.mesh_size = mesh_size
                self.surface_loop = surface_loop
                self.volume = volume

        return Box(x0, x1, y0, y1, z0, z1, surface_loop, vol, mesh_size=mesh_size)

    def add_torus(
        self,
        irad,
        orad,
        mesh_size=None,
        R=numpy.eye(3),
        x0=numpy.array([0.0, 0.0, 0.0]),
        variant="extrude_lines",
    ):

        if variant == "extrude_lines":
            return self._add_torus_extrude_lines(
                irad, orad, mesh_size=mesh_size, R=R, x0=x0
            )
        assert variant == "extrude_circle"
        return self._add_torus_extrude_circle(
            irad, orad, mesh_size=mesh_size, R=R, x0=x0
        )

    def _add_torus_extrude_lines(
        self,
        irad,
        orad,
        mesh_size=None,
        R=numpy.eye(3),
        x0=numpy.array([0.0, 0.0, 0.0]),
    ):
        """Create Gmsh code for the torus in the x-y plane under the coordinate
        transformation

        .. math::
            \\hat{x} = R x + x_0.

        :param irad: inner radius of the torus
        :param orad: outer radius of the torus
        """
        # Add circle
        x0t = numpy.dot(R, numpy.array([0.0, orad, 0.0]))
        # Get circles in y-z plane
        Rc = numpy.array([[0.0, 0.0, 1.0], [0.0, 1.0, 0.0], [1.0, 0.0, 0.0]])
        c = self.add_circle(x0 + x0t, irad, mesh_size=mesh_size, R=numpy.dot(R, Rc))

        rot_axis = [0.0, 0.0, 1.0]
        rot_axis = numpy.dot(R, rot_axis)
        point_on_rot_axis = [0.0, 0.0, 0.0]
        point_on_rot_axis = numpy.dot(R, point_on_rot_axis) + x0

        # Form the torus by extruding the circle three times by 2/3*pi. This
        # works around the inability of Gmsh to extrude by pi or more. The
        # Extrude() macro returns an array; the first [0] entry in the array is
        # the entity that has been extruded at the far end. This can be used
        # for the following Extrude() step.  The second [1] entry of the array
        # is the surface that was created by the extrusion.
        previous = c.curve_loop.curves
        angle = 2 * numpy.pi / 3
        all_surfaces = []
        for i in range(3):
            for k, p in enumerate(previous):
                # ts1[] = Extrude {{0,0,1}, {0,0,0}, 2*Pi/3}{Line{tc1};};
                # ...
                top, surf, _ = self.revolve(
                    p,
                    rotation_axis=rot_axis,
                    point_on_axis=point_on_rot_axis,
                    angle=angle,
                )
                all_surfaces.append(surf)
                previous[k] = top

        # compound_surface = CompoundSurface(all_surfaces)

        surface_loop = self.add_surface_loop(all_surfaces)
        vol = self.add_volume(surface_loop)
        return vol

    def _add_torus_extrude_circle(
        self,
        irad,
        orad,
        mesh_size=None,
        R=numpy.eye(3),
        x0=numpy.array([0.0, 0.0, 0.0]),
    ):
        """Create Gmsh code for the torus under the coordinate transformation

        .. math::
            \\hat{x} = R x + x_0.

        :param irad: inner radius of the torus
        :param orad: outer radius of the torus
        """
        # Add circle
        x0t = numpy.dot(R, numpy.array([0.0, orad, 0.0]))
        Rc = numpy.array([[0.0, 0.0, 1.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
        c = self.add_circle(x0 + x0t, irad, mesh_size=mesh_size, R=numpy.dot(R, Rc))

        rot_axis = [0.0, 0.0, 1.0]
        rot_axis = numpy.dot(R, rot_axis)
        point_on_rot_axis = [0.0, 0.0, 0.0]
        point_on_rot_axis = numpy.dot(R, point_on_rot_axis) + x0

        # Form the torus by extruding the circle three times by 2/3*pi. This
        # works around the inability of Gmsh to extrude by pi or more. The
        # Extrude() macro returns an array; the first [0] entry in the array is
        # the entity that has been extruded at the far end. This can be used
        # for the following Extrude() step.  The second [1] entry of the array
        # is the surface that was created by the extrusion. The third [2-end]
        # is a list of all the planes of the lateral surface.
        previous = c.plane_surface
        all_volumes = []
        num_steps = 3
        for _ in range(num_steps):
            top, vol, _ = self.revolve(
                previous,
                rotation_axis=rot_axis,
                point_on_axis=point_on_rot_axis,
                angle=2 * numpy.pi / num_steps,
            )
            previous = top
            all_volumes.append(vol)

        assert int(gmsh.__version__.split(".")[0])
        self._COMPOUND_ENTITIES.append((3, [v._ID for v in all_volumes]))

    def add_pipe(
        self,
        outer_radius,
        inner_radius,
        length,
        R=numpy.eye(3),
        x0=numpy.array([0.0, 0.0, 0.0]),
        mesh_size=None,
        variant="rectangle_rotation",
    ):
        if variant == "rectangle_rotation":
            return self._add_pipe_by_rectangle_rotation(
                outer_radius, inner_radius, length, R=R, x0=x0, mesh_size=mesh_size
            )
        assert variant == "circle_extrusion"
        return self._add_pipe_by_circle_extrusion(
            outer_radius, inner_radius, length, R=R, x0=x0, mesh_size=mesh_size
        )

    def _add_pipe_by_rectangle_rotation(
        self,
        outer_radius,
        inner_radius,
        length,
        R=numpy.eye(3),
        x0=numpy.array([0.0, 0.0, 0.0]),
        mesh_size=None,
    ):
        """Hollow cylinder.
        Define a rectangle, extrude it by rotation.
        """
        X = numpy.array(
            [
                [0.0, outer_radius, -0.5 * length],
                [0.0, outer_radius, +0.5 * length],
                [0.0, inner_radius, +0.5 * length],
                [0.0, inner_radius, -0.5 * length],
            ]
        )
        # Apply transformation.
        X = [numpy.dot(R, x) + x0 for x in X]
        # Create points set.
        p = [self.add_point(x, mesh_size=mesh_size) for x in X]

        # Define edges.
        e = [
            self.add_line(p[0], p[1]),
            self.add_line(p[1], p[2]),
            self.add_line(p[2], p[3]),
            self.add_line(p[3], p[0]),
        ]

        rot_axis = [0.0, 0.0, 1.0]
        rot_axis = numpy.dot(R, rot_axis)
        point_on_rot_axis = [0.0, 0.0, 0.0]
        point_on_rot_axis = numpy.dot(R, point_on_rot_axis) + x0

        # Extrude all edges three times by 2*Pi/3.
        previous = e
        angle = 2 * numpy.pi / 3
        all_surfaces = []
        # com = []
        for i in range(3):
            for k, p in enumerate(previous):
                # ts1[] = Extrude {{0,0,1}, {0,0,0}, 2*Pi/3}{Line{tc1};};
                top, surf, _ = self.revolve(
                    p,
                    rotation_axis=rot_axis,
                    point_on_axis=point_on_rot_axis,
                    angle=angle,
                )
                # if k==0:
                #     com.append(surf)
                # else:
                #     all_names.appends(surf)
                all_surfaces.append(surf)
                previous[k] = top
        #
        # cs = CompoundSurface(com)
        # Now just add surface loop and volume.
        # all_surfaces = all_names + [cs]
        surface_loop = self.add_surface_loop(all_surfaces)
        vol = self.add_volume(surface_loop)
        return vol

    def _add_pipe_by_circle_extrusion(
        self,
        outer_radius,
        inner_radius,
        length,
        R=numpy.eye(3),
        x0=numpy.array([0.0, 0.0, 0.0]),
        mesh_size=None,
    ):
        """Hollow cylinder.
        Define a ring, extrude it by translation.
        """
        # Define ring which to Extrude by translation.
        Rc = numpy.array([[0.0, 0.0, 1.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
        c_inner = self.add_circle(
            x0,
            inner_radius,
            mesh_size=mesh_size,
            R=numpy.dot(R, Rc),
            make_surface=False,
        )
        circ = self.add_circle(
            x0,
            outer_radius,
            mesh_size=mesh_size,
            R=numpy.dot(R, Rc),
            holes=[c_inner.curve_loop],
        )

        # Now Extrude the ring surface.
        _, vol, _ = self.extrude(
            circ.plane_surface, translation_axis=numpy.dot(R, [length, 0, 0])
        )
        return vol

    def in_surface(self, input_entity, surface):
        """Embed the point(s) or curve(s) in the given surface. The surface mesh will
        conform to the mesh of the point(s) or curves(s).
        """
        self._EMBED_QUEUE.append((input_entity, surface))

    def in_volume(self, input_entity, volume):
        """Embed the point(s)/curve(s)/surface(s) in the given volume. The volume mesh
        will conform to the mesh of the input entities.
        """
        self._EMBED_QUEUE.append((input_entity, volume))


class BoundaryLayer:
    def __init__(
        self,
        lcmin,
        lcmax,
        distmin,
        distmax,
        edges_list=None,
        faces_list=None,
        nodes_list=None,
    ):
        self.lcmin = lcmin
        self.lcmax = lcmax
        self.distmin = distmin
        self.distmax = distmax
        # Don't use [] as default argument, cf.
        # <https://stackoverflow.com/a/113198/353337>
        self.edges_list = edges_list if edges_list else []
        self.faces_list = faces_list if faces_list else []
        self.nodes_list = nodes_list if nodes_list else []

    def exec(self):
        tag1 = gmsh.model.mesh.field.add("Distance")

        if self.edges_list:
            gmsh.model.mesh.field.setNumbers(
                tag1, "EdgesList", [e._ID for e in self.edges_list]
            )
            # edge nodes must be specified, too, cf.
            # <https://gitlab.onelab.info/gmsh/gmsh/-/issues/812#note_9454>
            nodes = list(set([p for e in self.edges_list for p in e.points]))
            gmsh.model.mesh.field.setNumbers(tag1, "NodesList", [n._ID for n in nodes])
        if self.faces_list:
            gmsh.model.mesh.field.setNumbers(
                tag1, "FacesList", [f._ID for f in self.faces_list]
            )
        if self.nodes_list:
            gmsh.model.mesh.field.setNumbers(
                tag1, "NodesList", [n._ID for n in self.nodes_list]
            )

        tag2 = gmsh.model.mesh.field.add("Threshold")
        gmsh.model.mesh.field.setNumber(tag2, "IField", tag1)
        gmsh.model.mesh.field.setNumber(tag2, "LcMin", self.lcmin)
        gmsh.model.mesh.field.setNumber(tag2, "LcMax", self.lcmax)
        gmsh.model.mesh.field.setNumber(tag2, "DistMin", self.distmin)
        gmsh.model.mesh.field.setNumber(tag2, "DistMax", self.distmax)
        self.tag = tag2


class SetBackgroundMesh:
    def __init__(self, fields, operator):
        self.fields = fields
        self.operator = operator

    def exec(self):
        tag = gmsh.model.mesh.field.add(self.operator)
        gmsh.model.mesh.field.setNumbers(
            tag, "FieldsList", [f.tag for f in self.fields]
        )
        gmsh.model.mesh.field.setAsBackgroundMesh(tag)



from math import pi

import gmsh


class Ball:
    """
    Creates a sphere.

    Parameters
    ----------
    center: array-like[3]
        Center of the ball.
    radius: float
        Radius of the ball.
    x0: float
        If specified and `x0 > -1`, the ball is cut off at `x0*radius`
        parallel to the y-z plane.
    x1: float
        If specified and `x1 < +1`, the ball is cut off at `x1*radius`
        parallel to the y-z plane.
    alpha: float
        If specified and `alpha < 2*pi`, the points between `alpha` and
        `2*pi` w.r.t. to the x-y plane are not part of the object.
    char_length: float
        If specified, sets the `Characteristic Length` property.
    """

    def __init__(self, center, radius, angle1=-pi / 2, angle2=pi / 2, angle3=2 * pi):
        self.center = center
        self.radius = radius
        self._ID = gmsh.model.occ.addSphere(
            *center, radius, angle1=angle1, angle2=angle2, angle3=angle3
        )
        self.dim_tag = (3, self._ID)
        self.dim_tags = [self.dim_tag]

    def __repr__(self):
        return f"<pygmsh Ball object (OCC), ID {self._ID}>"



class CurveLoop:
    """
    Increments the Line ID every time a new object is created that inherits
    from LineBase.

    Parameters
    ----------
    id0 : str
        If no unique ID is given, the object global is incremented.
    lines : list
        Containing the lines defining the shape.

    Notes
    -----
    A line loop must be a closed loop, and the elementary lines should be ordered and
    oriented (negating to specify reverse orientation). If the orientation is correct,
    but the ordering is wrong, Gmsh will actually reorder the list internally to create
    a consistent loop.
    """

    dim = 1

    def __init__(self, env, curves):
        for k in range(len(curves) - 1):
            assert curves[k].points[-1] == curves[k + 1].points[0]
        assert curves[-1].points[-1] == curves[0].points[0]
        self._ID = env.addCurveLoop([c._ID for c in curves])

        self.curves = curves

    def __len__(self):
        return len(self.curves)

    def __repr__(self):
        curves = ", ".join([str(l._ID) for l in self.curves])
        return f"<pygmsh CurveLoop object, ID {self._ID}, curves ({curves})>"



class Volume:
    """
    Creates a volume.

    Parameters
    ----------
    surface_loop : list
        Contain the identification numbers of all the surface
        loops defining the volume.
    holes : list
        List containing surface loop objects that represents polygon holes.

    Notes
    -----
    The first surface loop defines the exterior boundary of the volume;
    all other surface loops define holes in the volume.

    A surface loop defining a hole should not have any surfaces in common
    with the exterior surface loop (in which case it is not a hole,
    and the two volumes should be defined separately).

    Likewise, a surface loop defining a hole should not have any surfaces
    in common with another surface loop defining a hole in the same volume
    (in which case the two surface loops should be combined).
    """

    dim = 3

    def __init__(self, env, surface_loop, holes=None):
        if holes is None:
            holes = []

        self.surface_loop = surface_loop
        self.holes = holes

        surface_loops = [surface_loop] + holes
        self._ID = env.addVolume([s._ID for s in surface_loops])



class Dummy:
    def __init__(self, dim, id0):
        assert isinstance(id0, int)
        self.dim = dim
        self._ID = id0
        self.dim_tag = (dim, id0)
        self.dim_tags = [self.dim_tag]

    def __repr__(self):
        return f"<pygmsh Dummy object, ID {self._ID}>"



from .line_base import LineBase
from .point import Point


class CircleArc(LineBase):
    """
    Creates a circle arc.

    Parameters
    ----------
    start : Point
        Coordinates of start point needed to construct circle-arc.
    center : Point
        Coordinates of center point needed to construct circle-arc.
    end : Point
        Coordinates of end point needed to construct circle-arc.
    """

    def __init__(self, env, start, center, end):
        assert isinstance(start, Point)
        assert isinstance(center, Point)
        assert isinstance(end, Point)
        id0 = env.addCircleArc(start._ID, center._ID, end._ID)
        self.dim_tags = [(1, id0)]
        super().__init__(id0, [start, center, end])



import pathlib

import exdown
import pytest

this_dir = pathlib.Path(__file__).resolve().parent


@pytest.mark.parametrize(
    "string, lineno",
    exdown.extract(this_dir.parent / "README.md", syntax_filter="python"),
)
def test_readme(string, lineno):
    exec(string)



import math

import gmsh

from .. import common
from .ball import Ball
from .box import Box
from .cone import Cone
from .cylinder import Cylinder
from .disk import Disk
from .dummy import Dummy
from .rectangle import Rectangle
from .torus import Torus
from .wedge import Wedge


class Geometry(common.CommonGeometry):
    def __init__(self):
        super().__init__(gmsh.model.occ)

    def __exit__(self, *a):
        # TODO remove once gmsh 4.7.0 is out
        # <https://gitlab.onelab.info/gmsh/gmsh/-/issues/1001>
        gmsh.option.setNumber("Mesh.CharacteristicLengthMin", 0.0)
        gmsh.option.setNumber("Mesh.CharacteristicLengthMax", 1.0e22)
        gmsh.finalize()

    @property
    def characteristic_length_min(self):
        return gmsh.option.getNumber("Mesh.CharacteristicLengthMin")

    @property
    def characteristic_length_max(self):
        return gmsh.option.getNumber("Mesh.CharacteristicLengthMax")

    @characteristic_length_min.setter
    def characteristic_length_min(self, val):
        gmsh.option.setNumber("Mesh.CharacteristicLengthMin", val)

    @characteristic_length_max.setter
    def characteristic_length_max(self, val):
        gmsh.option.setNumber("Mesh.CharacteristicLengthMax", val)

    def force_outward_normals(self, tag):
        self._OUTWARD_NORMALS.append(tag)

    def revolve(self, *args, **kwargs):
        if len(args) >= 4:
            angle = args[3]
        else:
            assert "angle" in kwargs
            angle = kwargs["angle"]

        assert angle < 2 * math.pi
        return super()._revolve(*args, **kwargs)

    def add_rectangle(self, *args, mesh_size=None, **kwargs):
        entity = Rectangle(*args, **kwargs)
        if mesh_size is not None:
            self._SIZE_QUEUE.append((entity, mesh_size))
        return entity

    def add_disk(self, *args, mesh_size=None, **kwargs):
        entity = Disk(*args, **kwargs)
        if mesh_size is not None:
            self._SIZE_QUEUE.append((entity, mesh_size))
        return entity

    def add_ball(self, *args, mesh_size=None, **kwargs):
        obj = Ball(*args, **kwargs)
        if mesh_size is not None:
            self._SIZE_QUEUE.append((obj, mesh_size))
        return obj

    def add_box(self, *args, mesh_size=None, **kwargs):
        box = Box(*args, **kwargs)
        if mesh_size is not None:
            self._SIZE_QUEUE.append((box, mesh_size))
        return box

    def add_cone(self, *args, mesh_size=None, **kwargs):
        cone = Cone(*args, **kwargs)
        if mesh_size is not None:
            self._SIZE_QUEUE.append((cone, mesh_size))
        return cone

    def add_cylinder(self, *args, mesh_size=None, **kwargs):
        cyl = Cylinder(*args, **kwargs)
        if mesh_size is not None:
            self._SIZE_QUEUE.append((cyl, mesh_size))
        return cyl

    def add_ellipsoid(self, center, radii, mesh_size=None):
        obj = Ball(center, 1.0)
        self.dilate(obj, center, radii)
        if mesh_size is not None:
            self._SIZE_QUEUE.append((obj, mesh_size))
        return obj

    def add_torus(self, *args, mesh_size=None, **kwargs):
        obj = Torus(*args, **kwargs)
        if mesh_size is not None:
            self._SIZE_QUEUE.append((obj, mesh_size))
        return obj

    def add_wedge(self, *args, mesh_size=None, **kwargs):
        obj = Wedge(*args, **kwargs)
        if mesh_size is not None:
            self._SIZE_QUEUE.append((obj, mesh_size))
        return obj

    def boolean_intersection(self, entities):
        """Boolean intersection, see
        https://gmsh.info/doc/texinfo/gmsh.html#Boolean-operations input_entity
        and tool_entity are called object and tool in gmsh documentation.
        """
        entities = [e if isinstance(e, list) else [e] for e in entities]

        ent = [e.dim_tag for e in entities[0]]
        # form subsequent intersections
        # https://gitlab.onelab.info/gmsh/gmsh/-/issues/999
        for e in entities[1:]:
            out, _ = gmsh.model.occ.intersect(
                ent,
                [ee.dim_tag for ee in e],
                removeObject=True,
                removeTool=True,
            )
            if len(out) == 0:
                raise RuntimeError("Empty intersection.")
            assert all(out[0] == item for item in out)
            ent = [out[0]]

        return [Dummy(*ent[0])]

    def boolean_union(self, entities):
        """Boolean union, see
        https://gmsh.info/doc/texinfo/gmsh.html#Boolean-operations input_entity
        and tool_entity are called object and tool in gmsh documentation.
        """
        entities = [e if isinstance(e, list) else [e] for e in entities]

        dim_tags, _ = gmsh.model.occ.fuse(
            [e.dim_tag for e in entities[0]],
            [ee.dim_tag for e in entities[1:] for ee in e],
            removeObject=True,
            removeTool=True,
        )
        return [Dummy(*dim_tag) for dim_tag in dim_tags]

    def boolean_difference(self, d0, d1, delete_first=True, delete_other=True):
        """Boolean difference, see
        https://gmsh.info/doc/texinfo/gmsh.html#Boolean-operations input_entity
        and tool_entity are called object and tool in gmsh documentation.
        """
        d0 = d0 if isinstance(d0, list) else [d0]
        d1 = d1 if isinstance(d1, list) else [d1]
        dim_tags, _ = gmsh.model.occ.cut(
            [d.dim_tag for d in d0],
            [d.dim_tag for d in d1],
            removeObject=delete_first,
            removeTool=delete_other,
        )
        return [Dummy(*dim_tag) for dim_tag in dim_tags]

    def boolean_fragments(self, d0, d1):
        """Boolean fragments, see
        https://gmsh.info/doc/texinfo/gmsh.html#Boolean-operations input_entity
        and tool_entity are called object and tool in gmsh documentation.
        """
        d0 = d0 if isinstance(d0, list) else [d0]
        d1 = d1 if isinstance(d1, list) else [d1]
        dim_tags, _ = gmsh.model.occ.fragment(
            [d.dim_tag for d in d0],
            [d.dim_tag for d in d1],
        )
        return [Dummy(*dim_tag) for dim_tag in dim_tags]



import gmsh


class Wedge:
    """
    Creates a right angular wedge.

    x0 : array-like[3]
        The 3 coordinates of the right-angle point.
    extends : array-like[3]
        List of the 3 extends of the box edges.
    top_extend : float
        Defines the top X extent.
    """

    def __init__(self, x0, extents, top_extent=None):
        self.x0 = x0
        self.extents = extents
        self.top_extent = top_extent

        self._ID = gmsh.model.occ.addWedge(*x0, *extents, ltx=top_extent)
        self.dim_tags = [(3, self._ID)]

    def __repr__(self):
        return f"<pygmsh Wedge object, ID {self._ID}>"



class SurfaceLoop:
    """
    Creates a surface loop (a shell).

    Parameters
    ----------
    surfaces : list
        Contain the identification numbers of all the elementary surfaces that
        constitute the surface loop.

    Notes
    -----
    A surface loop must always represent a closed shell, and the elementary surfaces
    should be oriented consistently (using negative identification numbers to specify
    reverse orientation).
    """

    dim = 2

    def __init__(self, env, surfaces):
        self.surfaces = surfaces
        self._ID = env.addSurfaceLoop([s._ID for s in surfaces])



from math import pi

import gmsh


class Torus:
    """
    Creates a torus.

    center : array-like[3]
        The 3 coordinates of its center.
    radius0 : float
        Inner radius.
    radius1 : float
        Outer radius.
    alpha : float
        Defines the angular opening.
    """

    def __init__(self, center, radius0, radius1, alpha=2 * pi):
        assert len(center) == 3

        self.center = center
        self.radius0 = radius0
        self.radius1 = radius1
        self.alpha = alpha

        self._ID = gmsh.model.occ.addTorus(*center, radius0, radius1, angle=alpha)
        self.dim_tag = (3, self._ID)
        self.dim_tags = [self.dim_tag]

    def __repr__(self):
        return f"<pygmsh Torus object, ID {self._ID}>"



import gmsh
import meshio
import numpy

from .helpers import extract_to_meshio


def optimize(mesh, method="", verbose=False):
    mesh.remove_lower_dimensional_cells()
    mesh.cell_data = {}

    # read into meshio like
    # <https://gitlab.onelab.info/gmsh/gmsh/-/blob/master/demos/api/import_perf.py>
    gmsh.initialize()
    # add dummy entity
    dim = 3
    tag = gmsh.model.addDiscreteEntity(dim=dim)
    #
    nodes = numpy.arange(1, len(mesh.points) + 1)
    assert mesh.points.shape[1] == 3
    gmsh.model.mesh.addNodes(dim, tag, nodes, mesh.points.flat)
    for cell_block in mesh.cells:
        gmsh.model.mesh.addElementsByType(
            tag,
            meshio.gmsh.meshio_to_gmsh_type[cell_block.type],
            [],
            cell_block.data.flatten() + 1,
        )
    gmsh.model.mesh.optimize(method, force=True)
    mesh = extract_to_meshio()
    gmsh.finalize()

    # This writes a temporary file and reads it into gmsh ("merge"). There are other
    # ways of feeding gmsh a mesh
    # (https://gitlab.onelab.info/gmsh/gmsh/-/issues/1030#note_11435), but let's not do
    # that for now.
    # with tempfile.TemporaryDirectory() as tmpdirname:
    #     tmpdir = Path(tmpdirname)
    #     tmpfile = tmpdir / "tmp.msh"
    #     mesh.write(tmpfile)
    #     gmsh.initialize()
    #     if verbose:
    #         gmsh.option.setNumber("General.Terminal", 1)
    #     gmsh.merge(str(tmpfile))
    #     # We need force=True because we're reading from a discrete mesh
    #     gmsh.model.mesh.optimize(method, force=True)
    #     mesh = extract_to_meshio()
    #     gmsh.finalize()
    return mesh


def print_stats(mesh):
    import termplotlib

    q = mesh.q_radius_ratio
    q_hist, q_bin_edges = numpy.histogram(
        q, bins=numpy.linspace(0.0, 1.0, num=41, endpoint=True)
    )

    grid = termplotlib.subplot_grid((1, 2), column_widths=None, border_style=None)
    grid[0, 0].hist(q_hist, q_bin_edges, bar_width=1, strip=True)
    grid[0, 1].aprint("min quality: {:5.3f}".format(numpy.min(q)))
    grid[0, 1].aprint("avg quality: {:5.3f}".format(numpy.average(q)))
    grid[0, 1].aprint("max quality: {:5.3f}".format(numpy.max(q)))

    grid.show()



import gmsh


class Box:
    """
    Creates a box.

    Parameters
    ----------
    x0 : array-like[3]
        List containing the x, y, z values of the start point.
    extends : array-like[3]
        List of the 3 extents of the box edges.
    char_length : float
        Characteristic length of the mesh elements of this polygon.
    """

    dim = 3

    def __init__(self, x0, extents, char_length=None):
        assert len(x0) == 3
        assert len(extents) == 3
        self.x0 = x0
        self.extents = extents
        self._ID = gmsh.model.occ.addBox(*x0, *extents)
        self.dim_tag = (3, self._ID)
        self.dim_tags = [self.dim_tag]



from .line_base import LineBase
from .point import Point


class BSpline(LineBase):
    """
    Creates a B-spline.

    Parameters
    ----------
    control_points : List[Point]
        Contains the identification numbers of the control points.
    """

    def __init__(self, env, control_points):
        for c in control_points:
            assert isinstance(c, Point)
        assert len(control_points) > 1

        id0 = env.addBSpline([c._ID for c in control_points])
        super().__init__(id0, control_points)



from .line_base import LineBase
from .point import Point


class Line(LineBase):
    """
    Creates a straight line segment.

    Parameters
    ----------
    p0 : Object
        Point object that represents the start of the line.
    p1 : Object
        Point object that represents the end of the line.

    Attributes
    ----------
    points : array-like[1][2]
        List containing the begin and end points of the line.
    """

    dim = 1

    def __init__(self, env, p0, p1):
        assert isinstance(p0, Point)
        assert isinstance(p1, Point)
        id0 = env.addLine(p0._ID, p1._ID)
        self.dim_tag = (1, id0)
        self.dim_tags = [self.dim_tag]
        super().__init__(id0, [p0, p1])

    def __repr__(self):
        pts = ", ".join(str(p._ID) for p in self.points)
        return f"<pygmsh Line object, ID {self._ID}, points ({pts})>"



import gmsh

__gmsh_version__ = gmsh.__version__

try:
    # Python 3.8
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata

try:
    __version__ = metadata.version("pygmsh")
except Exception:
    __version__ = "unknown"



import argparse
from sys import version_info

import meshio

from .__about__ import __gmsh_version__, __version__
from ._optimize import optimize


def optimize_cli(argv=None):
    parser = argparse.ArgumentParser(
        description=("Optimize mesh."),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument("infile", type=str, help="mesh to optimize")
    parser.add_argument("outfile", type=str, help="optimized mesh")

    parser.add_argument(
        "-q",
        "--quiet",
        dest="verbose",
        action="store_false",
        default=True,
        help="suppress output",
    )

    parser.add_argument(
        "-m",
        "--method",
        default="",
        # Valid choices are on
        # https://gmsh.info/doc/texinfo/gmsh.html#Namespace-gmsh_002fmodel_002fmesh
        help='method (e.g., "", Netgen, ...)',
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=_get_version_text(),
        help="display version information",
    )
    args = parser.parse_args(argv)

    mesh = meshio.read(args.infile)
    optimize(mesh, method=args.method, verbose=args.verbose).write(args.outfile)


def _get_version_text():
    return "\n".join(
        [
            f"pygmsh {__version__} "
            f"[Gmsh {__gmsh_version__}, "
            f"Python {version_info.major}.{version_info.minor}.{version_info.micro}]",
            "Copyright (c) 2013-2020 Nico Schlömer et al.",
        ]
    )



import copy


class LineBase:
    """
    Parameters
    ----------
    id0 : int
    points: list of int
    """

    dim = 1

    def __init__(self, id0, points):
        self._ID = id0
        self.points = points

    def __neg__(self):
        neg_self = copy.deepcopy(self)
        neg_self._ID = -self._ID
        neg_self.points = self.points[::-1]
        return neg_self



from .line_base import LineBase
from .point import Point


class EllipseArc(LineBase):
    """
    Creates an ellipse arc.

    Parameters
    ----------
    start : Point
        Coordinates of start point needed to construct elliptic arc.
    center : Point
        Coordinates of center point needed to construct elliptic arc.
    point_on_major_axis : Point
        Point on the center axis of ellipse.
    end : Point
        Coordinates of end point needed to construct elliptic arc.
    """

    def __init__(self, env, start, center, point_on_major_axis, end):
        assert isinstance(start, Point)
        assert isinstance(center, Point)
        assert isinstance(point_on_major_axis, Point)
        assert isinstance(end, Point)

        id0 = env.addEllipseArc(start._ID, center._ID, point_on_major_axis._ID, end._ID)
        super().__init__(id0, [start, center, end])

        self.points = [start, center, end]
        self.point_on_major_axis = point_on_major_axis

    def __repr__(self):
        pts = ", ".join(str(p._ID) for p in self.points)
        return f"<pygmsh EllipseArc object, ID {self._ID}, points ({pts})>"



import numpy


class Polygon:
    dim = 2

    def __init__(self, host, points, mesh_size=None, holes=None, make_surface=True):
        if holes is None:
            holes = []
        else:
            assert make_surface

        if isinstance(mesh_size, list):
            assert len(points) == len(mesh_size)
        else:
            mesh_size = len(points) * [mesh_size]

        points = numpy.asarray(points)
        if points.shape[1] == 2:
            points = numpy.column_stack([points, numpy.zeros_like(points[:, 0])])

        # Create points.
        self.points = [
            host.add_point(x, mesh_size=l) for x, l in zip(points, mesh_size)
        ]
        # Create lines
        self.curves = [
            host.add_line(self.points[k], self.points[k + 1])
            for k in range(len(self.points) - 1)
        ] + [host.add_line(self.points[-1], self.points[0])]

        self.lines = self.curves

        self.curve_loop = host.add_curve_loop(self.curves)
        # self.surface = host.add_plane_surface(ll, holes) if make_surface else None
        if make_surface:
            self.surface = host.add_plane_surface(self.curve_loop, holes)
            self.dim_tag = self.surface.dim_tag
            self.dim_tags = self.surface.dim_tags
            self._ID = self.surface._ID

    def __repr__(self):
        return "<pygmsh Polygon object>"



"""Tests module for helpers in tests."""
import numpy as np
from helpers import compute_volume

import pygmsh


def test():
    with pygmsh.geo.Geometry() as geom:
        geom.add_circle([0, 0, 0], 1, 0.1, make_surface=False)
        mesh = geom.generate_mesh()
        ref = 2 * np.pi
        assert np.abs(compute_volume(mesh) - ref) < 1e-2 * ref


if __name__ == "__main__":
    test()
